#!/usr/bin/env pypy
# -*- coding: utf-8 -*-

from __future__ import print_function
import re, sys, time
from itertools import count
from collections import namedtuple

###############################################################################
# Piece-Square tables. Tune these to change sunfish's behaviour
###############################################################################

piece = { 'P': 100, 'N': 280, 'B': 320, 'R': 479, 'Q': 929, 'K': 60000 }
pst = {
    'P': (   0,   0,   0,   0,   0,   0,   0,   0,
            78,  83,  86,  73, 102,  82,  85,  90,
             7,  29,  21,  44,  40,  31,  44,   7,
           -17,  16,  -2,  15,  14,   0,  15, -13,
           -26,   3,  10,   9,   6,   1,   0, -23,
           -22,   9,   5, -11, -10,  -2,   3, -19,
           -31,   8,  -7, -37, -36, -14,   3, -31,
             0,   0,   0,   0,   0,   0,   0,   0),
    'N': ( -66, -53, -75, -75, -10, -55, -58, -70,
            -3,  -6, 100, -36,   4,  62,  -4, -14,
            10,  67,   1,  74,  73,  27,  62,  -2,
            24,  24,  45,  37,  33,  41,  25,  17,
            -1,   5,  31,  21,  22,  35,   2,   0,
           -18,  10,  13,  22,  18,  15,  11, -14,
           -23, -15,   2,   0,   2,   0, -23, -20,
           -74, -23, -26, -24, -19, -35, -22, -69),
    'B': ( -59, -78, -82, -76, -23,-107, -37, -50,
           -11,  20,  35, -42, -39,  31,   2, -22,
            -9,  39, -32,  41,  52, -10,  28, -14,
            25,  17,  20,  34,  26,  25,  15,  10,
            13,  10,  17,  23,  17,  16,   0,   7,
            14,  25,  24,  15,   8,  25,  20,  15,
            19,  20,  11,   6,   7,   6,  20,  16,
            -7,   2, -15, -12, -14, -15, -10, -10),
    'R': (  35,  29,  33,   4,  37,  33,  56,  50,
            55,  29,  56,  67,  55,  62,  34,  60,
            19,  35,  28,  33,  45,  27,  25,  15,
             0,   5,  16,  13,  18,  -4,  -9,  -6,
           -28, -35, -16, -21, -13, -29, -46, -30,
           -42, -28, -42, -25, -25, -35, -26, -46,
           -53, -38, -31, -26, -29, -43, -44, -53,
           -30, -24, -18,   5,  -2, -18, -31, -32),
    'Q': (   6,   1,  -8,-104,  69,  24,  88,  26,
            14,  32,  60, -10,  20,  76,  57,  24,
            -2,  43,  32,  60,  72,  63,  43,   2,
             1, -16,  22,  17,  25,  20, -13,  -6,
           -14, -15,  -2,  -5,  -1, -10, -20, -22,
           -30,  -6, -13, -11, -16, -11, -16, -27,
           -36, -18,   0, -19, -15, -15, -21, -38,
           -39, -30, -31, -13, -31, -36, -34, -42),
    'K': (   4,  54,  47, -99, -99,  60,  83, -62,
           -32,  10,  55,  56,  56,  55,  10,   3,
           -62,  12, -57,  44, -67,  28,  37, -31,
           -55,  50,  11,  -4, -19,  13,   0, -49,
           -55, -43, -52, -28, -51, -47,  -8, -50,
           -47, -42, -43, -79, -64, -32, -29, -32,
            -4,   3, -14, -50, -57, -18,  13,   4,
            17,  30,  -3, -14,   6,  -1,  40,  18),
}
# Pad tables and join piece and pst dictionaries
for k, table in pst.items():
    padrow = lambda row: (0,) + tuple(x+piece[k] for x in row) + (0,)
    pst[k] = sum((padrow(table[i*8:i*8+8]) for i in range(8)), ())
    pst[k] = (0,)*20 + pst[k] + (0,)*20

###############################################################################
# Global constants
###############################################################################

# Our board is represented as a 120 character string. The padding allows for
# fast detection of moves that don't stay within the board.
A1, H1, A8, H8 = 91, 98, 21, 28
initial = (
    '         \n'  #   0 -  9
    '         \n'  #  10 - 19
    ' rnbqkbnr\n'  #  20 - 29
    ' pppppppp\n'  #  30 - 39
    ' ........\n'  #  40 - 49
    ' ........\n'  #  50 - 59
    ' ........\n'  #  60 - 69
    ' ........\n'  #  70 - 79
    ' PPPPPPPP\n'  #  80 - 89
    ' RNBQKBNR\n'  #  90 - 99
    '         \n'  # 100 -109
    '         \n'  # 110 -119
)

# Lists of possible moves for each piece type.
N, E, S, W = -10, 1, 10, -1
directions = {
    'P': (N, N+N, N+W, N+E),
    'N': (N+N+E, E+N+E, E+S+E, S+S+E, S+S+W, W+S+W, W+N+W, N+N+W),
    'B': (N+E, S+E, S+W, N+W),
    'R': (N, E, S, W),
    'Q': (N, E, S, W, N+E, S+E, S+W, N+W),
    'K': (N, E, S, W, N+E, S+E, S+W, N+W)
}

# Mate value must be greater than 8*queen + 2*(rook+knight+bishop)
# King value is set to twice this value such that if the opponent is
# 8 queens up, but we got the king, we still exceed MATE_VALUE.
# When a MATE is detected, we'll set the score to MATE_UPPER - plies to get there
# E.g. Mate in 3 will be MATE_UPPER - 6
MATE_LOWER = piece['K'] - 10*piece['Q']
MATE_UPPER = piece['K'] + 10*piece['Q']

# The table size is the maximum number of elements in the transposition table.
TABLE_SIZE = 1e7

# Constants for tuning search
QS_LIMIT = 219
EVAL_ROUGHNESS = 13
DRAW_TEST = True


###############################################################################
# Chess logic
###############################################################################

class Position(namedtuple('Position', 'board score wc bc ep kp')):
    """ A state of a chess game
    board -- a 120 char representation of the board
    score -- the board evaluation
    wc -- the castling rights, [west/queen side, east/king side]
    bc -- the opponent castling rights, [west/king side, east/queen side]
    ep - the en passant square
    kp - the king passant square
    """

    def gen_moves(self):
        # For each of our pieces, iterate through each possible 'ray' of moves,
        # as defined in the 'directions' map. The rays are broken e.g. by
        # captures or immediately in case of pieces such as knights.
        for i, p in enumerate(self.board):
            if not p.isupper(): continue
            for d in directions[p]:
                for j in count(i+d, d):
                    q = self.board[j]
                    # Stay inside the board, and off friendly pieces
                    if q.isspace() or q.isupper(): break
                    # Pawn move, double move and capture
                    if p == 'P' and d in (N, N+N) and q != '.': break
                    if p == 'P' and d == N+N and (i < A1+N or self.board[i+N] != '.'): break
                    if p == 'P' and d in (N+W, N+E) and q == '.' \
                            and j not in (self.ep, self.kp, self.kp-1, self.kp+1): break
                    # Move it
                    yield (i, j)
                    # Stop crawlers from sliding, and sliding after captures
                    if p in 'PNK' or q.islower(): break
                    # Castling, by sliding the rook next to the king
                    if i == A1 and self.board[j+E] == 'K' and self.wc[0]: yield (j+E, j+W)
                    if i == H1 and self.board[j+W] == 'K' and self.wc[1]: yield (j+W, j+E)

    def rotate(self):
        ''' Rotates the board, preserving enpassant '''
        return Position(
            self.board[::-1].swapcase(), -self.score, self.bc, self.wc,
            119-self.ep if self.ep else 0,
            119-self.kp if self.kp else 0)

    def nullmove(self):
        ''' Like rotate, but clears ep and kp '''
        return Position(
            self.board[::-1].swapcase(), -self.score,
            self.bc, self.wc, 0, 0)

    def move(self, move):
        i, j = move
        p, q = self.board[i], self.board[j]
        put = lambda board, i, p: board[:i] + p + board[i+1:]
        # Copy variables and reset ep and kp
        board = self.board
        wc, bc, ep, kp = self.wc, self.bc, 0, 0
        score = self.score + self.value(move)
        # Actual move
        board = put(board, j, board[i])
        board = put(board, i, '.')
        # Castling rights, we move the rook or capture the opponent's
        if i == A1: wc = (False, wc[1])
        if i == H1: wc = (wc[0], False)
        if j == A8: bc = (bc[0], False)
        if j == H8: bc = (False, bc[1])
        # Castling
        if p == 'K':
            wc = (False, False)
            if abs(j-i) == 2:
                kp = (i+j)//2
                board = put(board, A1 if j < i else H1, '.')
                board = put(board, kp, 'R')
        # Pawn promotion, double move and en passant capture
        if p == 'P':
            if A8 <= j <= H8:
                board = put(board, j, 'Q')
            if j - i == 2*N:
                ep = i + N
            if j == self.ep:
                board = put(board, j+S, '.')
        # We rotate the returned position, so it's ready for the next player
        return Position(board, score, wc, bc, ep, kp).rotate()

    def value(self, move):
        i, j = move
        p, q = self.board[i], self.board[j]
        # Actual move
        score = pst[p][j] - pst[p][i]
        # Capture
        if q.islower():
            score += pst[q.upper()][119-j]
        # Castling check detection
        if abs(j-self.kp) < 2:
            score += pst['K'][119-j]
        # Castling
        if p == 'K' and abs(i-j) == 2:
            score += pst['R'][(i+j)//2]
            score -= pst['R'][A1 if j < i else H1]
        # Special pawn stuff
        if p == 'P':
            if A8 <= j <= H8:
                score += pst['Q'][j] - pst['P'][j]
            if j == self.ep:
                score += pst['P'][119-(j+S)]
        return score

###############################################################################
# Search logic
###############################################################################

# lower <= s(pos) <= upper
Entry = namedtuple('Entry', 'lower upper')

class Searcher:
    def __init__(self):
        self.tp_score = {}
        self.tp_move = {}
        self.history = set()
        self.nodes = 0

    def bound(self, pos, gamma, depth, root=True):
        """ returns r where
                s(pos) <= r < gamma    if gamma > s(pos)
                gamma <= r <= s(pos)   if gamma <= s(pos)"""
        self.nodes += 1

        # Depth <= 0 is QSearch. Here any position is searched as deeply as is needed for
        # calmness, and from this point on there is no difference in behaviour depending on
        # depth, so so there is no reason to keep different depths in the transposition table.
        depth = max(depth, 0)

        # Sunfish is a king-capture engine, so we should always check if we
        # still have a king. Notice since this is the only termination check,
        # the remaining code has to be comfortable with being mated, stalemated
        # or able to capture the opponent king.
        if pos.score <= -MATE_LOWER:
            return -MATE_UPPER

        # We detect 3-fold captures by comparing against previously
        # _actually played_ positions.
        # Note that we need to do this before we look in the table, as the
        # position may have been previously reached with a different score.
        # This is what prevents a search instability.
        # FIXME: This is not true, since other positions will be affected by
        # the new values for all the drawn positions.
        if DRAW_TEST:
            if not root and pos in self.history:
                return 0

        # Look in the table if we have already searched this position before.
        # We also need to be sure, that the stored search was over the same
        # nodes as the current search.
        entry = self.tp_score.get((pos, depth, root), Entry(-MATE_UPPER, MATE_UPPER))
        if entry.lower >= gamma and (not root or self.tp_move.get(pos) is not None):
            return entry.lower
        if entry.upper < gamma:
            return entry.upper

        # Here extensions may be added
        # Such as 'if in_check: depth += 1'

        # Generator of moves to search in order.
        # This allows us to define the moves, but only calculate them if needed.
        def moves():
            # First try not moving at all. We only do this if there is at least one major
            # piece left on the board, since otherwise zugzwangs are too dangerous.
            if depth > 0 and not root and any(c in pos.board for c in 'RBNQ'):
                yield None, -self.bound(pos.nullmove(), 1-gamma, depth-3, root=False)
            # For QSearch we have a different kind of null-move, namely we can just stop
            # and not capture anything else.
            if depth == 0:
                yield None, pos.score
            # Then killer move. We search it twice, but the tp will fix things for us.
            # Note, we don't have to check for legality, since we've already done it
            # before. Also note that in QS the killer must be a capture, otherwise we
            # will be non deterministic.
            killer = self.tp_move.get(pos)
            if killer and (depth > 0 or pos.value(killer) >= QS_LIMIT):
                yield killer, -self.bound(pos.move(killer), 1-gamma, depth-1, root=False)
            # Then all the other moves
            for move in sorted(pos.gen_moves(), key=pos.value, reverse=True):
            #for val, move in sorted(((pos.value(move), move) for move in pos.gen_moves()), reverse=True):
                # If depth == 0 we only try moves with high intrinsic score (captures and
                # promotions). Otherwise we do all moves.
                if depth > 0 or pos.value(move) >= QS_LIMIT:
                    yield move, -self.bound(pos.move(move), 1-gamma, depth-1, root=False)

        # Run through the moves, shortcutting when possible
        best = -MATE_UPPER
        for move, score in moves():
            best = max(best, score)
            if best >= gamma:
                # Clear before setting, so we always have a value
                if len(self.tp_move) > TABLE_SIZE: self.tp_move.clear()
                # Save the move for pv construction and killer heuristic
                self.tp_move[pos] = move
                break

        # Stalemate checking is a bit tricky: Say we failed low, because
        # we can't (legally) move and so the (real) score is -infty.
        # At the next depth we are allowed to just return r, -infty <= r < gamma,
        # which is normally fine.
        # However, what if gamma = -10 and we don't have any legal moves?
        # Then the score is actaully a draw and we should fail high!
        # Thus, if best < gamma and best < 0 we need to double check what we are doing.
        # This doesn't prevent sunfish from making a move that results in stalemate,
        # but only if depth == 1, so that's probably fair enough.
        # (Btw, at depth 1 we can also mate without realizing.)
        if best < gamma and best < 0 and depth > 0:
            is_dead = lambda pos: any(pos.value(m) >= MATE_LOWER for m in pos.gen_moves())
            if all(is_dead(pos.move(m)) for m in pos.gen_moves()):
                in_check = is_dead(pos.nullmove())
                best = -MATE_UPPER if in_check else 0

        # Clear before setting, so we always have a value
        if len(self.tp_score) > TABLE_SIZE: self.tp_score.clear()
        # Table part 2
        if best >= gamma:
            self.tp_score[pos, depth, root] = Entry(best, entry.upper)
        if best < gamma:
            self.tp_score[pos, depth, root] = Entry(entry.lower, best)

        return best

    def search(self, pos, history=()):
        """ Iterative deepening MTD-bi search """
        self.nodes = 0
        if DRAW_TEST:
            self.history = set(history)
            # print('# Clearing table due to new history')
            self.tp_score.clear()

        # In finished games, we could potentially go far enough to cause a recursion
        # limit exception. Hence we bound the ply.
        for depth in range(1, 1000):
            # The inner loop is a binary search on the score of the position.
            # Inv: lower <= score <= upper
            # 'while lower != upper' would work, but play tests show a margin of 20 plays
            # better.
            lower, upper = -MATE_UPPER, MATE_UPPER
            while lower < upper - EVAL_ROUGHNESS:
                gamma = (lower+upper+1)//2
                score = self.bound(pos, gamma, depth)
                if score >= gamma:
                    lower = score
                if score < gamma:
                    upper = score
            # We want to make sure the move to play hasn't been kicked out of the table,
            # So we make another call that must always fail high and thus produce a move.
            self.bound(pos, lower, depth)
            # If the game hasn't finished we can retrieve our move from the
            # transposition table.
            yield depth, self.tp_move.get(pos), self.tp_score.get((pos, depth, True)).lower


###############################################################################
# User interface
###############################################################################

# Python 2 compatability
if sys.version_info[0] == 2:
    input = raw_input


def parse(c):
    fil, rank = ord(c[0]) - ord('a'), int(c[1]) - 1
    return A1 + fil - 10*rank


def render(i):
    rank, fil = divmod(i - A1, 10)
    return chr(fil + ord('a')) + str(-rank + 1)


def print_pos(pos):
    print()
    uni_pieces = {'R':'♜', 'N':'♞', 'B':'♝', 'Q':'♛', 'K':'♚', 'P':'♟',
                  'r':'♖', 'n':'♘', 'b':'♗', 'q':'♕', 'k':'♔', 'p':'♙', '.':'·'}
    for i, row in enumerate(pos.board.split()):
        print(' ', 8-i, ' '.join(uni_pieces.get(p, p) for p in row))
    print('    a b c d e f g h \n\n')


def main():
    hist = [Position(initial, 0, (True,True), (True,True), 0, 0)]
    searcher = Searcher()
    while True:
        print_pos(hist[-1])

        if hist[-1].score <= -MATE_LOWER:
            print("You lost")
            break

        # We query the user until she enters a (pseudo) legal move.
        move = None
        while move not in hist[-1].gen_moves():
            match = re.match('([a-h][1-8])'*2, input('Your move: '))
            if match:
                move = parse(match.group(1)), parse(match.group(2))
            else:
                # Inform the user when invalid input (e.g. "help") is entered
                print("Please enter a move like g8f6")
        hist.append(hist[-1].move(move))

        # After our move we rotate the board and print it again.
        # This allows us to see the effect of our move.
        print_pos(hist[-1].rotate())

        if hist[-1].score <= -MATE_LOWER:
            print("You won")
            break

        # Fire up the engine to look for a move.
        start = time.time()
        for _depth, move, score in searcher.search(hist[-1], hist):
            if time.time() - start > 1:
                break

        if score == MATE_UPPER:
            print("Checkmate!")

        # The black player moves from a rotated position, so we have to
        # 'back rotate' the move before printing it.
        print("My move:", render(119-move[0]) + render(119-move[1]))
        hist.append(hist[-1].move(move))


if __name__ == '__main__':
    main()



from datetime import date
from pathlib import Path

import chess
import pytest

from ch0 import cli


def make_game(
    *,
    player_color: chess.Color = chess.WHITE,
    engine_kind: str = "random",
    engine_name: str = "random",
    book_path: str | None = None,
    book_chance: float = 0.0,
) -> cli.Game:
    return cli.Game(
        engine_kind=engine_kind,
        engine_name=engine_name,
        player_color=player_color,
        engine=None,
        verbose=False,
        book_path=book_path,
        book_chance=book_chance,
    )


def force_choice(monkeypatch: pytest.MonkeyPatch, move: chess.Move) -> None:
    monkeypatch.setattr(cli.random, "choice", lambda _: move)


def test_bot_makes_a_move_updates_pgn_and_turn(monkeypatch: pytest.MonkeyPatch) -> None:
    game = make_game()
    move = chess.Move.from_uci("e2e4")
    force_choice(monkeypatch, move)

    cli.bot_makes_a_move(game)

    assert game.board.peek() == move
    assert game.count == 1
    assert game.turn == chess.BLACK
    assert "\n1. e4" in game.pgn_text


def test_bot_makes_a_move_appends_black_move(monkeypatch: pytest.MonkeyPatch) -> None:
    game = make_game()
    game.board.push_san("e4")
    game.turn = chess.BLACK
    game.count = 1
    game.pgn_text = "\n1. e4"
    move = chess.Move.from_uci("e7e5")
    force_choice(monkeypatch, move)

    cli.bot_makes_a_move(game)

    assert game.count == 1
    assert game.pgn_text.endswith(" e5")


def test_bot_makes_a_move_sets_draw(monkeypatch: pytest.MonkeyPatch) -> None:
    game = make_game()
    game.board = chess.Board("7k/8/8/8/8/8/8/7K w - - 0 1")
    move = next(iter(game.board.legal_moves))
    force_choice(monkeypatch, move)

    cli.bot_makes_a_move(game)

    assert game.ended is True
    assert "The game is a draw" in game.pgn_text


def test_bot_makes_a_move_sets_checkmate(monkeypatch: pytest.MonkeyPatch) -> None:
    game = make_game(player_color=chess.WHITE)
    game.board = chess.Board("7k/5Q2/7K/8/8/8/8/8 w - - 0 1")
    mate_move = chess.Move.from_uci("f7g7")
    assert any(mate_move == move for move in game.board.legal_moves)
    force_choice(monkeypatch, mate_move)

    cli.bot_makes_a_move(game)

    assert game.ended is True
    assert game.board.is_checkmate()
    assert "black wins by checkmate" in game.pgn_text


def test_is_a_draw_stalemate() -> None:
    board = chess.Board("7k/5Q2/6K1/8/8/8/8/8 b - - 0 1")
    is_draw, reason = cli.is_a_draw(board)
    assert is_draw is True
    assert reason == "Stalemate"


def test_is_a_draw_insufficient_material() -> None:
    board = chess.Board("7k/8/8/8/8/8/8/7K w - - 0 1")
    is_draw, reason = cli.is_a_draw(board)
    assert is_draw is True
    assert reason == "Insufficient Material"


def test_is_a_draw_fifty_move_rule() -> None:
    board = chess.Board("7k/8/8/8/8/8/8/R6K w - - 100 1")
    is_draw, reason = cli.is_a_draw(board)
    assert is_draw is True
    assert reason == "Fifty-move rule"


def test_is_a_draw_threefold_repetition() -> None:
    board = chess.Board()
    repeat = ["g1f3", "g8f6", "f3g1", "f6g8"]
    for _ in range(2):
        for uci in repeat:
            board.push(chess.Move.from_uci(uci))

    is_draw, reason = cli.is_a_draw(board)
    assert is_draw is True
    assert reason == "Threefold repetition"


def test_finalize_pgn_headers() -> None:
    game_pgn = cli.finalize_pgn("1. e4 e5", chess.WHITE, "sunfish")
    headers = game_pgn.headers

    assert headers["Event"] == "Blind-chess match"
    assert headers["Site"] == "Terminal"
    assert headers["White"] == "Me"
    assert headers["Black"] == "sunfish Bot"
    assert headers["Date"] == date.today().isoformat()


def test_parse_command_and_slugify() -> None:
    assert cli.parse_command(":show") == "show"
    assert cli.parse_command("  MoVes ") == "moves"
    assert cli.parse_command("") is None

    assert cli._slugify_filename("My Bot 2!") == "my_bot_2"
    assert cli._slugify_filename("   ") == "bot"


def test_book_status_line(tmp_path: Path) -> None:
    assert cli._book_status_line(None, 0.5) == "Opening book: none"
    assert cli._book_status_line("missing.bin", 0.5) == "Opening book: none"

    book = tmp_path / "book.bin"
    book.write_bytes(b"")
    assert cli._book_status_line(str(book), 0.5) == f"Opening book: {book} (50%)"


def test_parse_player_move_rejects_non_capture_with_x() -> None:
    board = chess.Board()
    moves = [
        "e4",
        "c6",
        "Nf3",
        "d5",
        "d3",
        "dxe4",
        "dxe4",
        "Qxd1",
        "Kxd1",
        "Bg4",
        "Be2",
        "e6",
        "Ne5",
        "Bxe2",
        "Kxe2",
        "Nf6",
        "f3",
        "Bd6",
        "Nc4",
        "Bc7",
        "a4",
        "O-O",
        "Be3",
        "Nbd7",
        "a5",
        "a6",
        "Nbd2",
    ]
    for san in moves:
        board.push_san(san)

    assert cli._parse_player_move(board, "Nxe5") is None


def test_parse_player_move_accepts_capture_with_x() -> None:
    board = chess.Board()
    board.push_san("e4")
    board.push_san("d5")
    move = cli._parse_player_move(board, "exd5")
    assert move is not None
    assert board.is_capture(move)


def test_parse_player_move_accepts_non_capture() -> None:
    board = chess.Board()
    move = cli._parse_player_move(board, "e4")
    assert move is not None
    assert board.is_capture(move) is False


def test_cpl_helpers() -> None:
    empty = {
        "inaccuracies": 0,
        "mistakes": 0,
        "blunders": 0,
        "cpl": 0,
        "moves": 0,
        "worst_loss": -1,
        "worst_san": "",
        "worst_move_number": 0,
        "worst_is_white": True,
    }
    assert cli._average_cpl(empty) == 0
    assert cli._format_worst_move(empty) == "Worst move: n/a"

    entry = {
        **empty,
        "moves": 1,
        "cpl": 120,
        "worst_loss": 120,
        "worst_san": "Nf3",
        "worst_move_number": 3,
    }
    assert cli._average_cpl(entry) == 120
    assert cli._format_worst_move(entry) == "Worst move: 3. Nf3 (120 CPL)"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import chess.engine
import json
import argparse
import random
import time
import sys
import asyncio
import pathlib
import logging
import math


parser = argparse.ArgumentParser()
parser.add_argument("-cmd", nargs="?", help="Command of (UCI) engine to use")
parser.add_argument("-conf", nargs="?", help="Location of engines.json file to use")
parser.add_argument("-name", nargs="?", help="Name of engine to use from conf")
parser.add_argument("-selfplay", action="store_true", help="Play against itself")
parser.add_argument("-debug", action="store_true", help="Enable debugging of engine")
parser.add_argument("-movetime", type=int, default=0, help="Movetime in ms")
parser.add_argument("-nodes", type=int, default=0, help="Maximum nodes")
parser.add_argument(
    "-pvs",
    nargs="?",
    const=3,
    default=0,
    type=int,
    help="Show Principal Variations (when mcts)",
)
parser.add_argument(
    "-fen",
    help="Start from given position",
    default="rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1",
)


async def load_engine_from_cmd(cmd, debug=False):
    _, engine = await chess.engine.popen_uci(cmd.split())
    if hasattr(engine, "debug"):
        engine.debug(debug)
    return engine


async def load_engine_from_conf(engine_args, name, debug=False):
    args = next(a for a in engine_args if a["name"] == name)
    curdir = str(pathlib.Path(__file__).parent)
    popen_args = {}
    if "workingDirectory" in args:
        popen_args["cwd"] = args["workingDirectory"].replace("$FILE", curdir)
    cmd = args["command"].split()
    if cmd[0] == "$PYTHON":
        cmd[0] = sys.executable
    if args["protocol"] == "uci":
        _, engine = await chess.engine.popen_uci(cmd, **popen_args)
    elif args["protocol"] == "xboard":
        _, engine = await chess.engine.popen_xboard(cmd, **popen_args)
    if hasattr(engine, "debug"):
        engine.debug(debug)
    await engine.configure(
        {opt["name"]: opt["value"] for opt in args.get("options", [])}
    )
    return engine


def get_user_move(board):
    # Get well-formated move
    move = None
    while move is None:
        san_option = random.choice([board.san(m) for m in board.legal_moves])
        uci_option = random.choice([m.uci() for m in board.legal_moves])
        uci = input(f"Your move (e.g. {san_option} or {uci_option}): ")
        if uci in ("quit", "exit"):
            return None

        for parse in (board.parse_san, chess.Move.from_uci):
            try:
                move = parse(uci)
            except ValueError:
                pass

    # Check legality
    if move not in board.legal_moves:
        print("Illegal move.")
        return get_user_move(board)

    return move


def get_user_color():
    color = ""
    while color not in ("white", "black"):
        color = input("Do you want to be white or black? ")
    return chess.WHITE if color == "white" else chess.BLACK


def print_unicode_board(board, perspective=chess.WHITE):
    """Prints the position from a given perspective."""
    sc, ec = "\x1b[0;30;107m", "\x1b[0m"
    for r in range(8) if perspective == chess.BLACK else range(7, -1, -1):
        line = [f"{sc} {r+1}"]
        for c in range(8) if perspective == chess.WHITE else range(7, -1, -1):
            color = "\x1b[48;5;255m" if (r + c) % 2 == 1 else "\x1b[48;5;253m"
            if board.move_stack:
                if board.move_stack[-1].to_square == 8 * r + c:
                    color = "\x1b[48;5;153m"
                elif board.move_stack[-1].from_square == 8 * r + c:
                    color = "\x1b[48;5;153m"
            piece = board.piece_at(8 * r + c)
            line.append(
                color + (chess.UNICODE_PIECE_SYMBOLS[piece.symbol()] if piece else " ")
            )
        print(" " + " ".join(line) + f" {sc} {ec}")
    if perspective == chess.WHITE:
        print(f" {sc}   a b c d e f g h  {ec}\n")
    else:
        print(f" {sc}   h g f e d c b a  {ec}\n")


async def get_engine_move(engine, board, limit, game_id, multipv, debug=False):
    # XBoard engine doesn't support multipv, and there python-chess doesn't support
    # getting the first PV while playing a game.
    if isinstance(engine, chess.engine.XBoardProtocol):
        play_result = await engine.play(board, limit, game=game_id)
        return play_result.move

    multipv = min(multipv, board.legal_moves.count())
    with await engine.analysis(
        board, limit, game=game_id, info=chess.engine.INFO_ALL, multipv=multipv or None
    ) as analysis:

        infos = [None for _ in range(multipv)]
        first = True
        async for new_info in analysis:
            # If multipv = 0 it means we don't want them at all,
            # but uci requires MultiPV to be at least 1.
            if multipv and "multipv" in new_info:
                infos[new_info["multipv"] - 1] = new_info

            # Parse optional arguments into a dict
            if debug and "string" in new_info:
                print(new_info["string"])

            if not debug and all(infos) and "score" in analysis.info:
                if not first:
                    # print('\n'*(multipv+1), end='')
                    print(f"\u001b[1A\u001b[K" * (multipv + 1), end="")
                else:
                    first = False

                info = analysis.info
                score = info["score"].relative
                score = (
                    f"Score: {score.score()}"
                    if score.score() is not None
                    else f"Mate in {score.mate()}"
                )
                print(
                    f'{score}, nodes: {info.get("nodes", "N/A")}, nps: {info.get("nps", "N/A")},'
                    f' time: {float(info.get("time", 0)):.1f}',
                    end="",
                )
                print()

                for info in infos:
                    if "pv" in info:
                        variation = board.variation_san(info["pv"][:10])
                    else:
                        variation = ""

                    if "score" in info:
                        score = info["score"].relative
                        score = (
                            math.tanh(score.score() / 600)
                            if score.score() is not None
                            else score.mate()
                        )
                        key, *val = info.get("string", "").split()
                        if key == "pv_nodes":
                            nodes = int(val[0])
                            rel = nodes / analysis.info["nodes"]
                            score_rel = f"({score:.2f}, {rel*100:.0f}%)"
                        else:
                            score_rel = f"({score:.2f})"
                    else:
                        score_rel = ""

                    # Something about N
                    print(f'{info["multipv"]}: {score_rel} {variation}')

        return analysis.info["pv"][0]


async def play(engine, board, selfplay, pvs, time_limit, debug=False):
    if not selfplay:
        user_color = get_user_color()
    else:
        user_color = chess.WHITE

    if not board:
        board = chess.Board()

    game_id = random.random()

    while not board.is_game_over():
        print_unicode_board(board, perspective=user_color)
        if not selfplay and user_color == board.turn:
            move = get_user_move(board)
            if move is None:
                return
        else:
            move = await get_engine_move(
                engine, board, time_limit, game_id, pvs, debug=debug
            )
            print(f" My move: {board.san(move)}")
        board.push(move)

    # Print status
    print_unicode_board(board, perspective=user_color)
    print("Result:", board.result())


async def main():
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)

    if not args.conf:
        if args.cmd:
            engine = await load_engine_from_cmd(args.cmd, debug=args.debug)
        else:
            path = pathlib.Path(__file__).parent / "engines.json"
            if not path.is_file():
                print("Unable to locate engines.json file.")
                return
            conf = json.load(path.open())
    else:
        if args.conf:
            conf = json.load(open(args.conf))
        else:
            path = pathlib.Path(__file__).parent / "engines.json"
            if not path.is_file():
                print("Unable to locate engines.json file.")
                return
            conf = json.load(path.open())
        engine = await load_engine_from_conf(conf, args.name, debug=args.debug)

    if "author" in engine.id:
        print(f"Playing against {engine.id['name']} by {engine.id['author']}.")
    else:
        print(f"Playing against {engine.id['name']}.")

    board = chess.Board(args.fen)

    if args.movetime:
        limit = chess.engine.Limit(time=args.movetime / 1000)
    elif args.nodes:
        limit = chess.engine.Limit(nodes=args.nodes)
    else:
        limit = chess.engine.Limit(
            white_clock=30, black_clock=30, white_inc=1, black_inc=1
        )

    try:
        await play(
            engine,
            board,
            selfplay=args.selfplay,
            pvs=args.pvs,
            time_limit=limit,
            debug=args.debug,
        )
    finally:
        print("\nGoodbye!")
        await engine.quit()


asyncio.set_event_loop_policy(chess.engine.EventLoopPolicy())
try:
    asyncio.run(main())
except KeyboardInterrupt:
    pass



from types import SimpleNamespace

import pytest

from ch0 import cli


def feed_inputs(monkeypatch: pytest.MonkeyPatch, inputs: list[str]) -> None:
    iterator = iter(inputs)
    monkeypatch.setattr("builtins.input", lambda _: next(iterator))


def test_ask_yes_no_defaults_and_retries(monkeypatch: pytest.MonkeyPatch) -> None:
    feed_inputs(monkeypatch, ["", "y"])
    assert cli.ask_yes_no("Confirm?") is False

    feed_inputs(monkeypatch, ["maybe", "Y"])
    assert cli.ask_yes_no("Confirm?") is True


def test_ask_pgn_action_defaults_and_retries(monkeypatch: pytest.MonkeyPatch) -> None:
    feed_inputs(monkeypatch, [""])
    assert cli.ask_pgn_action() == "none"

    feed_inputs(monkeypatch, ["wat", "s"])
    assert cli.ask_pgn_action() == "save"


def test_ask_analysis_report_defaults_and_retries(monkeypatch: pytest.MonkeyPatch) -> None:
    feed_inputs(monkeypatch, [""])
    assert cli.ask_analysis_report() == "none"

    feed_inputs(monkeypatch, ["nope", "full"])
    assert cli.ask_analysis_report() == "full"


def test_engine_display_name_falls_back_to_command() -> None:
    engine = SimpleNamespace(id={"name": "Stockfish"})
    assert cli._engine_display_name("stockfish", engine) == "Stockfish"

    unnamed = SimpleNamespace(id={})
    assert cli._engine_display_name("/usr/bin/fish", unnamed) == "fish"


def test_spawn_uci_engine_handles_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    def boom(*_args, **_kwargs):
        raise FileNotFoundError

    monkeypatch.setattr(cli.chess.engine.SimpleEngine, "popen_uci", boom)
    assert cli._spawn_uci_engine("missing-engine") is None


def test_main_start_resign_quit(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]) -> None:
    feed_inputs(
        monkeypatch,
        [
            "start",
            "1",
            "1",
            "resign",
            "",
            "",
            "quit",
        ],
    )

    cli.main([])

    out = capsys.readouterr().out
    assert "Resigned." in out
    assert "Goodbye." in out


def test_main_lobby_version(monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]) -> None:
    feed_inputs(monkeypatch, ["version", "quit"])
    monkeypatch.setattr(cli, "_format_version", lambda: "ch0 9.9.9")

    cli.main([])

    out = capsys.readouterr().out
    assert "ch0 9.9.9" in out
    assert "Goodbye." in out



from typing import Dict, List, Any
import chess
import sys
import time
from .evaluate import evaluate_board, move_value, check_end_game

debug_info: Dict[str, Any] = {}


MATE_SCORE = 1000000000
MATE_THRESHOLD = 999000000


def next_move(depth: int, board: chess.Board, debug=True) -> chess.Move:
    """
    What is the next best move?
    """
    debug_info.clear()
    debug_info["nodes"] = 0
    t0 = time.time()

    move = minimax_root(depth, board)

    debug_info["time"] = time.time() - t0
    if debug == True:
        print(f"info {debug_info}")
    return move


def get_ordered_moves(board: chess.Board) -> List[chess.Move]:
    """
    Get legal moves.
    Attempt to sort moves by best to worst.
    Use piece values (and positional gains/losses) to weight captures.
    """
    end_game = check_end_game(board)

    def orderer(move):
        return move_value(board, move, end_game)

    in_order = sorted(
        board.legal_moves, key=orderer, reverse=(board.turn == chess.WHITE)
    )
    return list(in_order)


def minimax_root(depth: int, board: chess.Board) -> chess.Move:
    """
    What is the highest value move per our evaluation function?
    """
    # White always wants to maximize (and black to minimize)
    # the board score according to evaluate_board()
    maximize = board.turn == chess.WHITE
    best_move = -float("inf")
    if not maximize:
        best_move = float("inf")

    moves = get_ordered_moves(board)
    best_move_found = moves[0]

    for move in moves:
        board.push(move)
        # Checking if draw can be claimed at this level, because the threefold repetition check
        # can be expensive. This should help the bot avoid a draw if it's not favorable
        # https://python-chess.readthedocs.io/en/latest/core.html#chess.Board.can_claim_draw
        if board.can_claim_draw():
            value = 0.0
        else:
            value = minimax(
                depth - 1, board, -float("inf"), float("inf"), not maximize
            )
        board.pop()
        if maximize and value >= best_move:
            best_move = value
            best_move_found = move
        elif not maximize and value <= best_move:
            best_move = value
            best_move_found = move

    return best_move_found


def minimax(
    depth: int,
    board: chess.Board,
    alpha: float,
    beta: float,
    is_maximising_player: bool,
) -> float:
    """
    Core minimax logic.
    https://en.wikipedia.org/wiki/Minimax
    """
    debug_info["nodes"] += 1

    if board.is_checkmate():
        # The previous move resulted in checkmate
        return -MATE_SCORE if is_maximising_player else MATE_SCORE
    # When the game is over and it's not a checkmate it's a draw
    # In this case, don't evaluate. Just return a neutral result: zero
    elif board.is_game_over():
        return 0

    if depth == 0:
        return evaluate_board(board)

    if is_maximising_player:
        best_move = -float("inf")
        moves = get_ordered_moves(board)
        for move in moves:
            board.push(move)
            curr_move = minimax(
                depth - 1, board, alpha, beta, not is_maximising_player
            )
            # Each ply after a checkmate is slower, so they get ranked slightly less
            # We want the fastest mate!
            if curr_move > MATE_THRESHOLD:
                curr_move -= 1
            elif curr_move < -MATE_THRESHOLD:
                curr_move += 1
            best_move = max(
                best_move,
                curr_move,
            )
            board.pop()
            alpha = max(alpha, best_move)
            if beta <= alpha:
                return best_move
        return best_move
    else:
        best_move = float("inf")
        moves = get_ordered_moves(board)
        for move in moves:
            board.push(move)
            curr_move = minimax(
                depth - 1, board, alpha, beta, not is_maximising_player
            )
            if curr_move > MATE_THRESHOLD:
                curr_move -= 1
            elif curr_move < -MATE_THRESHOLD:
                curr_move += 1
            best_move = min(
                best_move,
                curr_move,
            )
            board.pop()
            beta = min(beta, best_move)
            if beta <= alpha:
                return best_move
        return best_move



import chess

# this module implement's Tomasz Michniewski's Simplified Evaluation Function
# https://www.chessprogramming.org/Simplified_Evaluation_Function
# note that the board layouts have been flipped and the top left square is A1

# fmt: off
piece_value = {
    chess.PAWN: 100,
    chess.ROOK: 500,
    chess.KNIGHT: 320,
    chess.BISHOP: 330,
    chess.QUEEN: 900,
    chess.KING: 20000
}

pawnEvalWhite = [
    0,  0,  0,  0,  0,  0,  0,  0,
    5, 10, 10, -20, -20, 10, 10,  5,
    5, -5, -10,  0,  0, -10, -5,  5,
    0,  0,  0, 20, 20,  0,  0,  0,
    5,  5, 10, 25, 25, 10,  5,  5,
    10, 10, 20, 30, 30, 20, 10, 10,
    50, 50, 50, 50, 50, 50, 50, 50,
    0, 0, 0, 0, 0, 0, 0, 0
]
pawnEvalBlack = list(reversed(pawnEvalWhite))

knightEval = [
    -50, -40, -30, -30, -30, -30, -40, -50,
    -40, -20, 0, 0, 0, 0, -20, -40,
    -30, 0, 10, 15, 15, 10, 0, -30,
    -30, 5, 15, 20, 20, 15, 5, -30,
    -30, 0, 15, 20, 20, 15, 0, -30,
    -30, 5, 10, 15, 15, 10, 5, -30,
    -40, -20, 0, 5, 5, 0, -20, -40,
    -50, -40, -30, -30, -30, -30, -40, -50
]

bishopEvalWhite = [
    -20, -10, -10, -10, -10, -10, -10, -20,
    -10, 5, 0, 0, 0, 0, 5, -10,
    -10, 10, 10, 10, 10, 10, 10, -10,
    -10, 0, 10, 10, 10, 10, 0, -10,
    -10, 5, 5, 10, 10, 5, 5, -10,
    -10, 0, 5, 10, 10, 5, 0, -10,
    -10, 0, 0, 0, 0, 0, 0, -10,
    -20, -10, -10, -10, -10, -10, -10, -20
]
bishopEvalBlack = list(reversed(bishopEvalWhite))

rookEvalWhite = [
    0, 0, 0, 5, 5, 0, 0, 0,
    -5, 0, 0, 0, 0, 0, 0, -5,
    -5, 0, 0, 0, 0, 0, 0, -5,
    -5, 0, 0, 0, 0, 0, 0, -5,
    -5, 0, 0, 0, 0, 0, 0, -5,
    -5, 0, 0, 0, 0, 0, 0, -5,
    5, 10, 10, 10, 10, 10, 10, 5,
    0, 0, 0, 0, 0, 0, 0, 0
]
rookEvalBlack = list(reversed(rookEvalWhite))

queenEval = [
    -20, -10, -10, -5, -5, -10, -10, -20,
    -10, 0, 0, 0, 0, 0, 0, -10,
    -10, 0, 5, 5, 5, 5, 0, -10,
    -5, 0, 5, 5, 5, 5, 0, -5,
    0, 0, 5, 5, 5, 5, 0, -5,
    -10, 5, 5, 5, 5, 5, 0, -10,
    -10, 0, 5, 0, 0, 0, 0, -10,
    -20, -10, -10, -5, -5, -10, -10, -20
]

kingEvalWhite = [
    20, 30, 10, 0, 0, 10, 30, 20,
    20, 20, 0, 0, 0, 0, 20, 20,
    -10, -20, -20, -20, -20, -20, -20, -10,
    20, -30, -30, -40, -40, -30, -30, -20,
    -30, -40, -40, -50, -50, -40, -40, -30,
    -30, -40, -40, -50, -50, -40, -40, -30,
    -30, -40, -40, -50, -50, -40, -40, -30,
    -30, -40, -40, -50, -50, -40, -40, -30
]
kingEvalBlack = list(reversed(kingEvalWhite))

kingEvalEndGameWhite = [
    50, -30, -30, -30, -30, -30, -30, -50,
    -30, -30,  0,  0,  0,  0, -30, -30,
    -30, -10, 20, 30, 30, 20, -10, -30,
    -30, -10, 30, 40, 40, 30, -10, -30,
    -30, -10, 30, 40, 40, 30, -10, -30,
    -30, -10, 20, 30, 30, 20, -10, -30,
    -30, -20, -10,  0,  0, -10, -20, -30,
    -50, -40, -30, -20, -20, -30, -40, -50
]
kingEvalEndGameBlack = list(reversed(kingEvalEndGameWhite))
# fmt: on


def move_value(board: chess.Board, move: chess.Move, endgame: bool) -> float:
    """
    How good is a move?
    A promotion is great.
    A weaker piece taking a stronger piece is good.
    A stronger piece taking a weaker piece is bad.
    Also consider the position change via piece-square table.
    """
    if move.promotion is not None:
        return -float("inf") if board.turn == chess.BLACK else float("inf")

    _piece = board.piece_at(move.from_square)
    if _piece:
        _from_value = evaluate_piece(_piece, move.from_square, endgame)
        _to_value = evaluate_piece(_piece, move.to_square, endgame)
        position_change = _to_value - _from_value
    else:
        raise Exception(f"A piece was expected at {move.from_square}")

    capture_value = 0.0
    if board.is_capture(move):
        capture_value = evaluate_capture(board, move)

    current_move_value = capture_value + position_change
    if board.turn == chess.BLACK:
        current_move_value = -current_move_value

    return current_move_value


def evaluate_capture(board: chess.Board, move: chess.Move) -> float:
    """
    Given a capturing move, weight the trade being made.
    """
    if board.is_en_passant(move):
        return piece_value[chess.PAWN]
    _to = board.piece_at(move.to_square)
    _from = board.piece_at(move.from_square)
    if _to is None or _from is None:
        raise Exception(
            f"Pieces were expected at _both_ {move.to_square} and {move.from_square}"
        )
    return piece_value[_to.piece_type] - piece_value[_from.piece_type]


def evaluate_piece(piece: chess.Piece, square: chess.Square, end_game: bool) -> int:
    piece_type = piece.piece_type
    mapping = []
    if piece_type == chess.PAWN:
        mapping = pawnEvalWhite if piece.color == chess.WHITE else pawnEvalBlack
    if piece_type == chess.KNIGHT:
        mapping = knightEval
    if piece_type == chess.BISHOP:
        mapping = bishopEvalWhite if piece.color == chess.WHITE else bishopEvalBlack
    if piece_type == chess.ROOK:
        mapping = rookEvalWhite if piece.color == chess.WHITE else rookEvalBlack
    if piece_type == chess.QUEEN:
        mapping = queenEval
    if piece_type == chess.KING:
        # use end game piece-square tables if neither side has a queen
        if end_game:
            mapping = (
                kingEvalEndGameWhite
                if piece.color == chess.WHITE
                else kingEvalEndGameBlack
            )
        else:
            mapping = kingEvalWhite if piece.color == chess.WHITE else kingEvalBlack

    return mapping[square]


def evaluate_board(board: chess.Board) -> float:
    """
    Evaluates the full board and determines which player is in a most favorable position.
    The sign indicates the side:
        (+) for white
        (-) for black
    The magnitude, how big of an advantage that player has
    """
    total = 0
    end_game = check_end_game(board)

    for square in chess.SQUARES:
        piece = board.piece_at(square)
        if not piece:
            continue

        value = piece_value[piece.piece_type] + evaluate_piece(piece, square, end_game)
        total += value if piece.color == chess.WHITE else -value

    return total


def check_end_game(board: chess.Board) -> bool:
    """
    Are we in the end game?
    Per Michniewski:
    - Both sides have no queens or
    - Every side which has a queen has additionally no other pieces or one minorpiece maximum.
    """
    queens = 0
    minors = 0

    for square in chess.SQUARES:
        piece = board.piece_at(square)
        if piece and piece.piece_type == chess.QUEEN:
            queens += 1
        if piece and (
            piece.piece_type == chess.BISHOP or piece.piece_type == chess.KNIGHT
        ):
            minors += 1

    if queens == 0 or (queens == 2 and minors <= 1):
        return True

    return False



#!/usr/bin/env -S uv run
import argparse
import io
import os
import random
import shlex
import subprocess
import tomllib
from datetime import date, datetime
from importlib import metadata
from pathlib import Path

import chess
import chess.pgn
import chess.polyglot
import chess.engine

from .engines.andoma.movegeneration import next_move as andoma_gen
from .engines.sunfish import sunfish_uci
from .engines.sunfish.tools import uci


# --- Colors (ANSI) ------------------------------------------------------------
# Works in most terminals. On Windows, ANSI is supported in modern terminals;
# if you need broader support, install `colorama` and it will be used if present.
try:
    import colorama  # type: ignore
    colorama.just_fix_windows_console()
except Exception:
    pass


class Style:
    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"

    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"


def c(text: str, *styles: str) -> str:
    return "".join(styles) + text + Style.RESET


# --- Version ----------------------------------------------------------------
def _get_version() -> str:
    for name in ("play-ch0", "ch0"):
        try:
            return metadata.version(name)
        except metadata.PackageNotFoundError:
            continue
        except Exception:
            break

    try:
        root = Path(__file__).resolve().parents[2]
        pyproject = root / "pyproject.toml"
        if pyproject.exists():
            data = tomllib.loads(pyproject.read_text(encoding="utf-8"))
            version = data.get("project", {}).get("version")
            if isinstance(version, str) and version:
                return version
    except Exception:
        pass

    return "unknown"


def _format_version() -> str:
    return f"ch0 {_get_version()}"


# --- Game logic --------------------------------------------------------------
class Game:
    def __init__(
        self,
        engine_kind: str,
        engine_name: str,
        player_color: chess.Color,
        engine: chess.engine.SimpleEngine | None = None,
        verbose: bool = False,
        book_path: str | None = None,
        book_chance: float = 0.5,
    ):
        self.board = chess.Board()
        self.engine_kind = engine_kind  # "random", "andoma", "sunfish", "uci"
        self.engine_name = engine_name
        self.player_color = player_color
        self.engine = engine
        self.verbose = verbose
        self.book_path = book_path
        self.book_chance = book_chance
        self.turn = chess.WHITE  # whose turn it is to move in our bookkeeping
        self.count = 0           # move number (full moves)
        self.pgn_text = ""
        self.ended = False

    def reset(self):
        self.board.set_fen(chess.STARTING_FEN)
        self.turn = chess.WHITE
        self.count = 0
        self.pgn_text = ""
        self.ended = False

    def close_engine(self):
        if self.engine is None:
            return
        try:
            self.engine.quit()
        except Exception:
            try:
                self.engine.close()
            except Exception:
                pass
        self.engine = None


def is_a_draw(board: chess.Board):
    """Return (is_draw, message)."""
    if board.is_stalemate():
        return True, "Stalemate"
    elif board.is_insufficient_material():
        return True, "Insufficient Material"
    elif board.can_claim_fifty_moves():
        return True, "Fifty-move rule"
    elif board.can_claim_threefold_repetition():
        return True, "Threefold repetition"
    return False, ""


def bool_color_to_string(color_b: chess.Color) -> str:
    return "white" if color_b == chess.WHITE else "black"


def finalize_pgn(pgn_str: str, player_color: chess.Color, engine_name: str):
    final_pgn = pgn_str + "\n\n"
    game_pgn = chess.pgn.read_game(io.StringIO(final_pgn))
    game_pgn.headers["Event"] = "Blind-chess match"
    game_pgn.headers["Site"] = "Terminal"
    if player_color == chess.WHITE:
        game_pgn.headers["White"] = "Me"
        game_pgn.headers["Black"] = f"{engine_name} Bot"
    else:
        game_pgn.headers["White"] = f"{engine_name} Bot"
        game_pgn.headers["Black"] = "Me"
    game_pgn.headers["Date"] = date.today().isoformat()
    return game_pgn


def bot_makes_a_move(game: Game):
    board = game.board
    move = random.choice(list(board.legal_moves))
    used_book = False

    if game.engine_kind == "andoma":
        move = andoma_gen(depth=4, board=board, debug=False)
    elif game.engine_kind == "sunfish":
        position = uci.from_fen(*board.fen().split())
        current_hist = (
            [position]
            if uci.get_color(position) == uci.WHITE
            else [position.rotate(), position]
        )
        total_time = random.randint(10, 60)
        _, uci_move_str = sunfish_uci.generate_move(current_hist, total_time)
        move = chess.Move.from_uci(uci_move_str)
    elif game.engine_kind == "uci":
        if game.engine is None:
            raise RuntimeError("UCI engine is not initialized.")
        think_time = random.uniform(0.1, 0.5)
        result = game.engine.play(board, chess.engine.Limit(time=think_time))
        move = result.move

    # optional opening book
    roll = random.random()
    if game.book_path and game.count < 15 and roll < game.book_chance:
        try:
            with chess.polyglot.open_reader(game.book_path) as reader:
                move = reader.weighted_choice(board).move
                used_book = True
        except (IndexError, FileNotFoundError):
            pass

    turn_number = board.fullmove_number
    move_san = board.san(move)
    board.push(move)

    if game.turn == chess.WHITE:
        game.count += 1
        game.pgn_text += f"\n{game.count}. {move_san}"
    else:
        game.pgn_text += f" {move_san}"

    # Minimal engine output (colored, no label)
    turn_prefix = c(f"{turn_number}.", Style.DIM)
    if game.verbose and used_book:
        print(
            f"{turn_prefix} {c(move_san, Style.MAGENTA, Style.BOLD)}"
            f"{c(' (book)', Style.DIM)}"
        )
    else:
        print(f"{turn_prefix} {c(move_san, Style.MAGENTA, Style.BOLD)}")

    # draw / checkmate handling
    check_draw, draw_type = is_a_draw(board)
    if check_draw:
        print(c(f"Draw: {draw_type}", Style.YELLOW, Style.BOLD))
        game.pgn_text += " { The game is a draw. } 1/2-1/2"
        game.ended = True
        return

    if board.is_checkmate():
        print(c("Checkmate.", Style.RED, Style.BOLD))
        result = "0-1" if game.player_color == chess.WHITE else "1-0"
        game.pgn_text += (
            f" {{ {bool_color_to_string(not game.player_color)} wins by checkmate. }} "
            f"{result}"
        )
        game.ended = True
        return

    game.turn = not game.turn


def _engine_display_name(command: str, engine: chess.engine.SimpleEngine) -> str:
    name = engine.id.get("name")
    if name:
        return name
    return os.path.basename(command.split()[0]) or "UCI"


def _spawn_uci_engine(command: str) -> chess.engine.SimpleEngine | None:
    cmd = shlex.split(command)
    if not cmd:
        return None
    try:
        return chess.engine.SimpleEngine.popen_uci(cmd, stderr=subprocess.DEVNULL)
    except (FileNotFoundError, PermissionError, OSError, chess.engine.EngineError):
        return None


def _book_status_line(book_path: str | None, book_chance: float) -> str:
    if book_path is None:
        return "Opening book: none"
    if os.path.exists(book_path):
        chance_pct = round(book_chance * 100)
        return f"Opening book: {book_path} ({chance_pct}%)"
    return "Opening book: none"


def choose_engine() -> tuple[str, str, chess.engine.SimpleEngine | None]:
    options = ["random", "andoma", "sunfish", "uci"]
    print(c("Choose engine:", Style.CYAN, Style.BOLD))
    for i, name in enumerate(options, start=1):
        print(f"  {c(str(i) + '.', Style.DIM)} {c(name, Style.CYAN)}")
    while True:
        choice = input(c("engine> ", Style.DIM)).strip().lower()
        if choice in {"1", "2", "3", "4"}:
            choice = options[int(choice) - 1]
        if choice in options:
            if choice != "uci":
                return choice, choice, None
            while True:
                cmd = input(c("uci engine path/command> ", Style.DIM)).strip()
                engine = _spawn_uci_engine(cmd)
                if engine is not None:
                    display_name = _engine_display_name(cmd, engine)
                    return "uci", display_name, engine
                print(c("Could not start engine. Try again.", Style.RED))
        print(c("Invalid choice.", Style.RED))


def choose_color():
    print(c("Choose your color:", Style.CYAN, Style.BOLD))
    print(f"  {c('1.', Style.DIM)} {c('white', Style.CYAN)}")
    print(f"  {c('2.', Style.DIM)} {c('black', Style.CYAN)}")
    print(f"  {c('3.', Style.DIM)} {c('random', Style.CYAN)}")
    while True:
        choice = input(c("color> ", Style.DIM)).strip().lower()
        if choice in {"1", "white"}:
            return chess.WHITE
        if choice in {"2", "black"}:
            return chess.BLACK
        if choice in {"3", "random"}:
            return random.choice([chess.WHITE, chess.BLACK])
        print(c("Invalid choice.", Style.RED))


def print_help():
    print(c("Lobby:", Style.CYAN, Style.BOLD))
    print(f"  {c('start', Style.CYAN)}  start a new game")
    print(f"  {c('quick', Style.CYAN)}  start with sunfish + random color")
    print(f"  {c('help', Style.CYAN)}   show this help")
    print(f"  {c('version', Style.CYAN)} show program version")
    print(f"  {c('quit', Style.CYAN)}   quit")
    print()
    print(c("In-game:", Style.CYAN, Style.BOLD))
    print(f"  {c('show', Style.CYAN)}   show the board")
    print(f"  {c('moves', Style.CYAN)}  show legal moves (SAN)")
    print(f"  {c('fen', Style.CYAN)}    show FEN")
    print(f"  {c('pgn', Style.CYAN)}    show PGN so far")
    print(f"  {c('resign', Style.CYAN)} resign the game")
    print()
    print(c("Or type a move in SAN, e.g. e4, Nf3, exd5, a8=Q.", Style.DIM))


def parse_command(s: str):
    """Normalize commands: ':show' and 'show' both become 'show'."""
    s = s.strip()
    if not s:
        return None
    if s.startswith(":"):
        s = s[1:]
    return s.lower()


def _parse_player_move(board: chess.Board, user_in: str) -> chess.Move | None:
    try:
        move = board.parse_san(user_in)
    except ValueError:
        return None
    if "x" in user_in and not board.is_capture(move):
        return None
    return move


def ask_yes_no(prompt: str, default_no: bool = True) -> bool:
    suffix = " [y/N]: " if default_no else " [Y/n]: "
    while True:
        ans = input(c(prompt + suffix, Style.DIM)).strip().lower()
        if not ans:
            return not default_no
        if ans in {"y", "yes"}:
            return True
        if ans in {"n", "no"}:
            return False
        print(c("Please answer y or n.", Style.RED))


def _slugify_filename(text: str) -> str:
    cleaned = []
    for ch in text.strip().lower():
        if ch.isalnum() or ch in {"-", "_"}:
            cleaned.append(ch)
        elif ch.isspace():
            cleaned.append("_")
    return "".join(cleaned) or "bot"


def ask_pgn_action() -> str:
    prompt = "Final PGN: (p)rint, (s)ave, or (n)one?"
    suffix = " [n]: "
    while True:
        ans = input(c(prompt + suffix, Style.DIM)).strip().lower()
        if not ans:
            return "none"
        if ans in {"p", "print"}:
            return "print"
        if ans in {"s", "save"}:
            return "save"
        if ans in {"n", "no", "none", "skip"}:
            return "none"
        print(c("Please answer print, save, or skip.", Style.RED))


def _score_to_cp(score: chess.engine.PovScore) -> int:
    return score.score(mate_score=100000)


def _summarize_cpl(loss_cp: int) -> str:
    return f"{loss_cp} CPL"


def _analyze_pgn_with_stockfish(
    game_pgn: chess.pgn.Game,
    stockfish_cmd: str,
) -> dict[chess.Color, dict[str, int | str | bool]] | None:
    cmd = stockfish_cmd.strip() or "stockfish"
    engine = _spawn_uci_engine(cmd)
    if engine is None:
        return None

    stats = {
        chess.WHITE: {
            "inaccuracies": 0,
            "mistakes": 0,
            "blunders": 0,
            "cpl": 0,
            "moves": 0,
            "worst_loss": -1,
            "worst_san": "",
            "worst_move_number": 0,
            "worst_is_white": True,
        },
        chess.BLACK: {
            "inaccuracies": 0,
            "mistakes": 0,
            "blunders": 0,
            "cpl": 0,
            "moves": 0,
            "worst_loss": -1,
            "worst_san": "",
            "worst_move_number": 0,
            "worst_is_white": False,
        },
    }
    board = game_pgn.board()
    limit = chess.engine.Limit(depth=12)

    try:
        for move in game_pgn.mainline_moves():
            mover = board.turn
            move_san = board.san(move)
            move_number = board.fullmove_number
            best_info = engine.analyse(board, limit)
            best_cp = _score_to_cp(best_info["score"].pov(mover))

            played_info = engine.analyse(board, limit, root_moves=[move])
            played_cp = _score_to_cp(played_info["score"].pov(mover))

            loss = max(0, best_cp - played_cp)
            stats[mover]["cpl"] += loss
            stats[mover]["moves"] += 1

            if loss > stats[mover]["worst_loss"]:
                stats[mover]["worst_loss"] = loss
                stats[mover]["worst_san"] = move_san
                stats[mover]["worst_move_number"] = move_number
                stats[mover]["worst_is_white"] = mover == chess.WHITE

            if loss >= 300:
                stats[mover]["blunders"] += 1
            elif loss >= 100:
                stats[mover]["mistakes"] += 1
            elif loss >= 50:
                stats[mover]["inaccuracies"] += 1

            board.push(move)
    finally:
        engine.quit()

    return stats


def ask_stockfish_command() -> str:
    prompt = "Stockfish command/path"
    suffix = " [stockfish]: "
    return input(c(prompt + suffix, Style.DIM)).strip()


def ask_analysis_report() -> str:
    prompt = "Post-game analysis: (s)tats, (a)ccuracy, (w)orst move, (f)ull, or (n)one?"
    suffix = " [n]: "
    while True:
        ans = input(c(prompt + suffix, Style.DIM)).strip().lower()
        if not ans:
            return "none"
        if ans in {"s", "stats"}:
            return "stats"
        if ans in {"a", "accuracy", "avg"}:
            return "accuracy"
        if ans in {"w", "worst"}:
            return "worst"
        if ans in {"f", "full", "report"}:
            return "full"
        if ans in {"n", "no", "none", "skip"}:
            return "none"
        print(c("Please answer stats, accuracy, worst, full, or none.", Style.RED))


def _average_cpl(stats_entry: dict[str, int | str | bool]) -> int:
    moves = int(stats_entry["moves"])
    if moves <= 0:
        return 0
    return round(int(stats_entry["cpl"]) / moves)


def _format_worst_move(stats_entry: dict[str, int | str | bool]) -> str:
    moves = int(stats_entry["moves"])
    if moves <= 0:
        return "Worst move: n/a"
    move_number = int(stats_entry["worst_move_number"])
    move_san = str(stats_entry["worst_san"])
    worst_loss = int(stats_entry["worst_loss"])
    prefix = f"{move_number}. " if bool(stats_entry["worst_is_white"]) else f"{move_number}... "
    return f"Worst move: {prefix}{move_san} ({_summarize_cpl(worst_loss)})"


def _parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=_format_version(),
        help="Show version and exit.",
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Show a subtle indicator when an opening book move is used.",
    )
    parser.add_argument(
        "--book",
        metavar="PATH",
        default="book.bin",
        help="Path to a Polyglot opening book (default: book.bin).",
    )
    parser.add_argument(
        "--book-chance",
        metavar="P",
        type=float,
        default=0.5,
        help="Chance to use an opening book move when available (0.0-1.0).",
    )
    return parser.parse_args(argv)


def main(argv: list[str] | None = None):
    args = _parse_args(argv)
    print(c("\nBlindfold Chess\n", Style.BOLD))
    print_help()
    print()

    game: Game | None = None

    while True:
        # If a game ended, optionally print PGN, then return to lobby.
        if game is not None and game.ended:
            if game.pgn_text:
                final_pgn = finalize_pgn(game.pgn_text, game.player_color, game.engine_name)
                action = ask_pgn_action()
                if action == "print":
                    print()
                    print(c("Final PGN:", Style.CYAN, Style.BOLD))
                    print(final_pgn)
                elif action == "save":
                    stamp = datetime.now().strftime("%Y%m%d_%H%M%S")
                    opponent = _slugify_filename(game.engine_name)
                    filename = f"ch0_{opponent}_{stamp}.pgn"
                    with open(filename, "w", encoding="utf-8") as handle:
                        handle.write(str(final_pgn))
                        handle.write("\n")
                    print(c(f"Saved PGN to {filename}", Style.DIM))

                analysis_choice = ask_analysis_report()
                if analysis_choice != "none":
                    stockfish_cmd = ask_stockfish_command()
                    stats = _analyze_pgn_with_stockfish(final_pgn, stockfish_cmd)
                    if stats is None:
                        print(c("Could not start Stockfish.", Style.RED))
                    else:
                        you_color = game.player_color
                        engine_color = not game.player_color
                        you_stats = stats[you_color]
                        engine_stats = stats[engine_color]
                        print()
                        print(c("Post-game analysis:", Style.CYAN, Style.BOLD))
                        if analysis_choice in {"stats", "full"}:
                            print(
                                f"  {c('You:', Style.CYAN)} "
                                f"{you_stats['inaccuracies']} inaccuracies, "
                                f"{you_stats['mistakes']} mistakes, "
                                f"{you_stats['blunders']} blunders"
                            )
                            print(
                                f"  {c(f'{game.engine_name} Bot:', Style.CYAN)} "
                                f"{engine_stats['inaccuracies']} inaccuracies, "
                                f"{engine_stats['mistakes']} mistakes, "
                                f"{engine_stats['blunders']} blunders"
                            )
                        if analysis_choice in {"accuracy", "full"}:
                            print(
                                f"  {c('You:', Style.CYAN)} Avg CPL {_average_cpl(you_stats)}"
                            )
                            print(
                                f"  {c(f'{game.engine_name} Bot:', Style.CYAN)} "
                                f"Avg CPL {_average_cpl(engine_stats)}"
                            )
                        if analysis_choice in {"worst", "full"}:
                            print(
                                f"  {c('You:', Style.CYAN)} "
                                f"{_format_worst_move(you_stats)}"
                            )
                            print(
                                f"  {c(f'{game.engine_name} Bot:', Style.CYAN)} "
                                f"{_format_worst_move(engine_stats)}"
                            )
            game.close_engine()
            game = None
            print()
            print(c("Lobby. Type 'start' or 'quick' to play.", Style.DIM))
            continue

        # No active game: only limited commands work.
        if game is None:
            user_in = input(c("> ", Style.DIM)).strip()
            if not user_in:
                continue

            cmd = parse_command(user_in)
            if cmd == "start":
                engine_kind, engine_name, engine = choose_engine()
                player_color = choose_color()
                game = Game(
                    engine_kind,
                    engine_name,
                    player_color,
                    engine=engine,
                    verbose=args.verbose,
                    book_path=args.book,
                    book_chance=args.book_chance,
                )

                print()
                print(
                    c("You:", Style.DIM)
                    + " "
                    + c(bool_color_to_string(player_color), Style.CYAN, Style.BOLD)
                    + c(" vs ", Style.DIM)
                    + c(engine_name, Style.CYAN, Style.BOLD)
                )
                print(c(_book_status_line(args.book, args.book_chance), Style.DIM))
                print(c("Tip: type 'show' to display the board.", Style.DIM))
                print()

                # If the engine is white, let it move first.
                if player_color == chess.BLACK:
                    bot_makes_a_move(game)
                continue

            if cmd == "quick":
                engine_kind = "sunfish"
                engine_name = "sunfish"
                engine = None
                player_color = random.choice([chess.WHITE, chess.BLACK])
                game = Game(
                    engine_kind,
                    engine_name,
                    player_color,
                    engine=engine,
                    verbose=args.verbose,
                    book_path=args.book,
                    book_chance=args.book_chance,
                )

                print()
                print(
                    c("You:", Style.DIM)
                    + " "
                    + c(bool_color_to_string(player_color), Style.CYAN, Style.BOLD)
                    + c(" vs ", Style.DIM)
                    + c(engine_name, Style.CYAN, Style.BOLD)
                )
                print(c(_book_status_line(args.book, args.book_chance), Style.DIM))
                print(c("Tip: type 'show' to display the board.", Style.DIM))
                print()

                if player_color == chess.BLACK:
                    bot_makes_a_move(game)
                continue

            if cmd == "help":
                print_help()
                print()
                continue
            if cmd == "version":
                print(_format_version())
                print()
                continue

            if cmd == "quit":
                print(c("Goodbye.", Style.DIM))
                if game is not None:
                    game.close_engine()
                break

            print(c("No active game. Type 'start' or 'quick' (or 'help', 'quit').", Style.RED))
            continue

        # If it's engine's turn, just let it move.
        if game.board.turn != game.player_color:
            bot_makes_a_move(game)
            continue

        turn_prompt = f"{game.board.fullmove_number}> "
        user_in = input(c(turn_prompt, Style.DIM)).strip()
        if not user_in:
            continue

        cmd = parse_command(user_in)

        # Known in-game commands
        if cmd in {"help", "show", "moves", "fen", "pgn", "resign", "quit", "start"}:
            if cmd == "help":
                print_help()
            elif cmd == "show":
                print(game.board)
            elif cmd == "moves":
                moves_in_uci = list(game.board.legal_moves)
                moves_in_san = [game.board.san(m) for m in moves_in_uci]
                print(" ".join(moves_in_san))
            elif cmd == "fen":
                print(game.board.fen())
            elif cmd == "pgn":
                print(finalize_pgn(game.pgn_text, game.player_color, game.engine_name))
            elif cmd == "resign":
                print(c("Resigned.", Style.YELLOW, Style.BOLD))
                result = "0-1" if game.player_color == chess.WHITE else "1-0"
                game.pgn_text += (
                    f" {{ {bool_color_to_string(game.player_color)} resigns. }} {result}"
                )
                game.ended = True
            elif cmd == "quit":
                print(c("Goodbye.", Style.DIM))
                game.close_engine()
                break
            elif cmd == "start":
                print(c("Game in progress. Finish or resign first.", Style.RED))
            continue

        # Otherwise, try to interpret it as a move in SAN
        move = _parse_player_move(game.board, user_in)
        if move is None:
            print(c("Illegal move / unknown command.", Style.RED))
            continue

        game.board.push(move)

        # Optional: extremely subtle acknowledgement (comment out if you want *zero* noise)
        # print(c("✓", Style.GREEN, Style.DIM))

        if game.turn == chess.WHITE:
            game.count += 1
            game.pgn_text += f"\n{game.count}. {user_in}"
        else:
            game.pgn_text += f" {user_in}"

        check_draw, draw_type = is_a_draw(game.board)
        if check_draw:
            print(c(f"Draw: {draw_type}", Style.YELLOW, Style.BOLD))
            game.pgn_text += " { The game is a draw. } 1/2-1/2"
            game.ended = True
            continue

        if game.board.is_checkmate():
            print(c("Checkmate. You win.", Style.GREEN, Style.BOLD))
            result = "1-0" if game.player_color == chess.WHITE else "0-1"
            game.pgn_text += (
                f" {{ {bool_color_to_string(game.player_color)} wins by checkmate. }} "
                f"{result}"
            )
            game.ended = True
            continue

        game.turn = not game.turn
        # engine will move in the next iteration


if __name__ == "__main__":
    main()
